# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import xml.etree.cElementTree as ET
from collections import OrderedDict

from driverng import MessageV2
from driverng import utils
from driverng import Worklist


logger = logging.getLogger('driverng.aqualinkxml')


class Driver(MessageV2):
    SPEC = {
        "code": "aqualinkxml",
        "label": "AQUALink"
    }
    ENCODING = "utf8"

    def parse(self, data):
        return ET.fromstring(data.encode(self.ENCODING))

    def to_string(self, message):
        return utils.prettify_xml(ET.tostring(message, 'utf-8'))

    def encode_worklist(self, worklist, context):
        msg = ""
        samples_orders = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""}
        )
        logger.info("samples orders %s detected", len(samples_orders))
        if len(samples_orders) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            for body in context["query"].findall("{http://schemas.xmlsoap.org/soap/envelope/}Body"):
                if body.find("{http://www.ngnydevices.tech/aqualis/3-0}ConveyorInitialization"):
                    logger.info("Response ConveyorInitialization detected")
                    msg = ET.Element("S:Envelope")
                    msg.set("xmlns:S", "http://schemas.xmlsoap.org/soap/envelope/")
                    body = ET.SubElement(msg, "S:Body")
                    conveyor = ET.SubElement(body, "ConveyorInitializationResponse")
                    conveyor.set("xmlns", "http://www.ngnydevices.tech/aqualis/3-0")
                    if len(samples_orders) == 0:
                        ET.SubElement(conveyor, "Result").text = "PrimaryTubeNotFound"
                    else:
                        ET.SubElement(conveyor, "Result").text = "Success"
                elif body.find("{http://www.ngnydevices.tech/aqualis/3-0}GetTests"):
                    rack_id = ""
                    hole_id = ""
                    tube_id = ""
                    logger.info("Response GetTests detected")
                    for get_tests in body.findall("{http://www.ngnydevices.tech/aqualis/3-0}GetTests"):
                        for primary_tube in get_tests.findall("{http://www.ngnydevices.tech/}PrimaryTube"):
                            if (primary_tube.findtext("{http://www.ngnydevices.tech/aqualis/3-0}Id")):
                                tube_id = primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text
                            else:
                                tube_id = ""
                            rack_id = primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location").find("{http://www.ngnydevices.tech/aqualis/3-0}RackId").text
                            hole_id = primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location").find("{http://www.ngnydevices.tech/aqualis/3-0}HoleId").text
                    msg = ET.Element("S:Envelope")
                    msg.set("xmlns:S", "http://schemas.xmlsoap.org/soap/envelope/")
                    body = ET.SubElement(msg, "S:Body")
                    bulk_order = ET.SubElement(body, "GetTestsResponse")
                    bulk_order.set("xmlns", "http://www.ngnydevices.tech/aqualis/3-0")
                    if not samples_orders:
                        logger.info("dem_auto_found %s", context.get("dem_auto_found"))
                        if context.get("dem_auto_found"):
                            ET.SubElement(bulk_order, "Result").text = "Success"
                        else:
                            ET.SubElement(bulk_order, "Result").text = "PrimaryTubeNotFound"
                        sample_xml = ET.SubElement(bulk_order, "PrimaryTube")
                        ET.SubElement(
                            sample_xml, "Id"
                        ).text = tube_id
                        location = ET.SubElement(sample_xml, "Location")
                        ET.SubElement(
                            location, "RackId"
                        ).text = ""
                        ET.SubElement(
                            location, "HoleId"
                        ).text = ""
                        tests = ET.SubElement(bulk_order, "Tests")
                        test = ET.SubElement(tests, "Test")
                        ET.SubElement(
                            test, "Id"
                        ).text = ""
                        ET.SubElement(
                            test, "Status"
                        ).text = ""
                    else:
                        ET.SubElement(bulk_order, "Result").text = "Success"
                    for sample, action, priority in worklist.get_worklist(
                        unique_together=["action", "priority"],
                        action={"del": worklist.IGNORE, worklist.ELSE: ""},
                        priority={"A": "STAT", "S": "STAT", worklist.ELSE: "Routine"},
                    ):
                        sample_xml = ET.SubElement(bulk_order, "PrimaryTube")
                        ET.SubElement(
                            sample_xml, "Id"
                        ).text = sample.sample_id
                        location = ET.SubElement(sample_xml, "Location")
                        ET.SubElement(
                            location, "RackId"
                        ).text = rack_id
                        ET.SubElement(
                            location, "HoleId"
                        ).text = hole_id
                        patient = ET.SubElement(bulk_order, "Patient")
                        ET.SubElement(
                            patient, "Id"
                        ).text = sample.patient_id
                        ET.SubElement(
                            patient, "FamilyName"
                        ).text = sample.last_name
                        ET.SubElement(
                            patient, "FirstName"
                        ).text = sample.first_name
                        ET.SubElement(
                            patient, "Sex"
                        ).text = sample.sex
                        ET.SubElement(
                            patient, "BirthDate"
                        ).text = utils.format_date(sample.birthdate, fmt="%d/%m/%Y")
                        sample_source = ET.SubElement(bulk_order, "SampleSource")
                        ET.SubElement(
                            sample_source, "Service"
                        ).text = sample.location
                        tests = ET.SubElement(bulk_order, "Tests")
                        order_status = OrderedDict()
                        for order in sample["orders"]:
                            order_status.setdefault(order.code, [])
                            order_status[order.code].append(order.dem_auto_test_status)
                        logger.info("order_status %s", order_status)
                        for code, status in order_status.items():
                            test = ET.SubElement(tests, "Test")
                            ET.SubElement(
                                test, "Id"
                            ).text = code
                            if min(status) == 0:
                                texte = "Pending"
                            else:
                                texte = "Done"
                            ET.SubElement(
                                test, "Status"
                            ).text = texte
                        # Si absence de tests, il faut envoyer quand même les balises vides...
                        if not sample["orders"]:
                            test = ET.SubElement(tests, "Test")
                            ET.SubElement(
                                test, "Id"
                            ).text = ""
                            ET.SubElement(
                                test, "Status"
                            ).text = ""
        return msg

    def decode_message(self, message):
        for body in message.findall("{http://schemas.xmlsoap.org/soap/envelope/}Body"):
            if body.find("{http://www.ngnydevices.tech/aqualis/3-0}GetTests"):
                worklist = Worklist(context="query")
                for get_tests in body.findall("{http://www.ngnydevices.tech/aqualis/3-0}GetTests"):
                    for primary_tube in get_tests.findall("{http://www.ngnydevices.tech/aqualis/3-0}PrimaryTube"):
                        rack = ""
                        loc = primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location")
                        rackid = loc.find("{http://www.ngnydevices.tech/aqualis/3-0}RackId")
                        holeid = loc.find("{http://www.ngnydevices.tech/aqualis/3-0}HoleId")
                        if rackid is not None and holeid is not None:
                            rack = "{}-{}".format(rackid.text.strip(), holeid.text.strip())
                        worklist.get_or_create(
                            type="patient",
                            sample_id=primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text.strip(),
                            rack=rack,
                            request_info="REQUEST_TESTS",
                        )
            elif body.find("{http://www.ngnydevices.tech/aqualis/3-0}ConveyorInitialization"):
                worklist = Worklist(context="query")
                for request in body.findall("{http://www.ngnydevices.tech/aqualis/3-0}ConveyorInitialization"):
                    for primary_tube in request.findall("{http://www.ngnydevices.tech/aqualis/3-0}Tube"):
                        worklist.get_or_create(
                            type="patient",
                            sample_id=primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text.strip(),
                            rack="{}-{}".format(
                                primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location").find("{http://www.ngnydevices.tech/aqualis/3-0}RackId").text.strip(),
                                primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location").find("{http://www.ngnydevices.tech/aqualis/3-0}HoleId").text.strip()
                            ),
                            request_info="REQUEST_INIT",
                        )
            elif body.find("{http://www.ngnydevices.tech/aqualis/3-0}SendResults"):
                worklist = Worklist(context="result")
                rack = ""
                for send_results in body.findall("{http://www.ngnydevices.tech/aqualis/3-0}SendResults"):
                    flags = []
                    sample_id = ""
                    for primary_tube in send_results.findall("{http://www.ngnydevices.tech/aqualis/3-0}ProcessedPrimaryTube"):
                        sample_id = primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text
                        rack = "{}-{}".format(
                            primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location").find("{http://www.ngnydevices.tech/aqualis/3-0}RackId").text.strip(),
                            primary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location").find("{http://www.ngnydevices.tech/aqualis/3-0}HoleId").text.strip()
                        )
                        for visual_analysis in primary_tube.findall("{http://www.ngnydevices.tech/aqualis/3-0}VisualAnalysis"):
                            # Width
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}Width"):
                                flags.append("Width:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}Width").text.strip()))

                            # Heigth
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}Height"):
                                flags.append("Height:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}Height").text.strip()))

                            # Volume estimation
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}VolumeEstimation"):
                                flags.append("Volume:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}VolumeEstimation").text.strip()))

                            # CapType
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}CapType"):
                                flags.append("CapType:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}CapType").text.strip()))

                            # Hémolyse
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}HValue"):
                                flags.append("Hemolyse:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}HValue").text.strip()))

                            # Ictère
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}IValue"):
                                flags.append("Ictere:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}IValue").text.strip()))

                            # Lactacte
                            if visual_analysis.findtext("{http://www.ngnydevices.tech/aqualis/3-0}LValue"):
                                flags.append("Lactate:{}".format(visual_analysis.find("{http://www.ngnydevices.tech/aqualis/3-0}LValue").text.strip()))

                            sample = worklist.get_or_create(
                                type="patient",
                                sample_id=sample_id.strip(),
                                rack=rack
                            )
                    test_secondary = ""
                    rackid_alicot = ""
                    holeid_alicot = ""
                    for generated_secondary_tube in send_results.findall("{http://www.ngnydevices.tech/aqualis/3-0}GeneratedSecondaryTubes"):
                        for secondary_tube in generated_secondary_tube.findall("{http://www.ngnydevices.tech/aqualis/3-0}SecondaryTube"):
                            for related_tests in secondary_tube.findall("{http://www.ngnydevices.tech/aqualis/3-0}RelatedTests"):
                                test_secondary = ""
                                for tests in related_tests.findall("{http://www.ngnydevices.tech/aqualis/3-0}Test"):
                                    test_secondary = tests.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text
                                    # Tube aliquoté
                                    if test_secondary.startswith("AL"):
                                        rackid_alicot = ""
                                        holeid_alicot = ""
                                        location = secondary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location")
                                        rackid_alicot = location.find("{http://www.ngnydevices.tech/aqualis/3-0}RackId")
                                        holeid_alicot = location.find("{http://www.ngnydevices.tech/aqualis/3-0}HoleId")
                    for test_results in send_results.findall("{http://www.ngnydevices.tech/aqualis/3-0}TestResults"):
                        for test in test_results.findall("{http://www.ngnydevices.tech/aqualis/3-0}Test"):
                            code = test.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text.strip()
                            flags_alicot = []
                            if (
                                code == test_secondary and
                                rackid_alicot is not None and
                                holeid_alicot is not None
                            ):
                                flags_alicot.extend(flags)
                                flags_alicot.append("{}-{}".format(
                                    rackid_alicot.text.strip(),
                                    holeid_alicot.text.strip())
                                )
                            sample.results.add(
                                code=code,
                                value=test.find("{http://www.ngnydevices.tech/aqualis/3-0}Status").text.strip(),
                                flags=flags_alicot if flags_alicot else flags
                            )

                    # Traitement du rack d'archive ou du rack de sérothèque
                    if rack and (rack.startswith("ARC") or rack.startswith("SER")):
                        sample.results.add(
                            code="RACK",
                            value=rack,
                        )

                    # Traitement du rack du tube secondaire du tube aliquoté
                    for generated_secondary_tube in send_results.findall("{http://www.ngnydevices.tech/aqualis/3-0}GeneratedSecondaryTubes"):
                        for secondary_tube in generated_secondary_tube.findall("{http://www.ngnydevices.tech/aqualis/3-0}SecondaryTube"):
                            for related_tests in secondary_tube.findall("{http://www.ngnydevices.tech/aqualis/3-0}RelatedTests"):
                                test_secondary = ""
                                for tests in related_tests.findall("{http://www.ngnydevices.tech/aqualis/3-0}Test"):
                                    test_secondary = tests.find("{http://www.ngnydevices.tech/aqualis/3-0}Id").text
                                    # Tube aliquoté
                                    if test_secondary.startswith("AL"):
                                        rack = ""
                                        rackid = ""
                                        holeid = ""
                                        location = secondary_tube.find("{http://www.ngnydevices.tech/aqualis/3-0}Location")
                                        rackid = location.find("{http://www.ngnydevices.tech/aqualis/3-0}RackId")
                                        holeid = location.find("{http://www.ngnydevices.tech/aqualis/3-0}HoleId")
                                        if rackid is not None and holeid is not None:
                                            rack = "{}-{}".format(rackid.text.strip(), holeid.text.strip())
                                            sample.results.add(
                                                code="RACK",
                                                value=rack,
                                            )
            else:
                raise NotImplementedError("Type de XML inconnu")
        return worklist
